#include <kernel.h>
#include <stdio.h>
#include "libpad.h"

static char padBuf[2][256] __attribute__((aligned(64)));

u32 old_pad[2];

void waitPadReady(int port, int slot) {
	int state;
	int lastState;
	char stateString[16];

	state = padGetState(port, slot);
	lastState = -1;
	while ((state != PAD_STATE_DISCONN) && (state != PAD_STATE_STABLE)
			&& (state != PAD_STATE_FINDCTP1)) {
		if (state != lastState) {
			padStateInt2String(state, stateString);
			printf("Please wait, pad(%d,%d) is in state %s\n", port, slot,
					stateString);
		}
		lastState = state;
		state=padGetState(port, slot);
		if (port == 1)
			break;
	}
	// Were the pad ever 'out of sync'?
	if (lastState != -1) {
		printf("Pad OK!\n");
	}
}

int initializePad(int port, int slot) {

	int ret;
	int modes;
	int i;

	waitPadReady(port, slot);

	// How many different modes can this device operate in?
	// i.e. get # entrys in the modetable
	modes = padInfoMode(port, slot, PAD_MODETABLE, -1);
	printf("The device has %d modes\n", modes);

	if (modes > 0) {
		printf("( ");
		for (i = 0; i < modes; i++) {
			printf("%d ", padInfoMode(port, slot, PAD_MODETABLE, i));
		}
		printf(")");
	}

	printf("It is currently using mode %d\n", padInfoMode(port, slot,
			PAD_MODECURID, 0));

	// If modes == 0, this is not a Dual shock controller
	// (it has no actuator engines)
	if (modes == 0) {
		printf("This is a digital controller?\n");
		return 1;
	}

	// Verify that the controller has a DUAL SHOCK mode
	i = 0;
	do {
		if (padInfoMode(port, slot, PAD_MODETABLE, i) == PAD_TYPE_DUALSHOCK)
			break;
		i++;
	} while (i < modes);
	if (i >= modes) {
		printf("This is no Dual Shock controller\n");
		return 1;
	}

	// If ExId != 0x0 => This controller has actuator engines
	// This check should always pass if the Dual Shock test above passed
	ret = padInfoMode(port, slot, PAD_MODECUREXID, 0);
	if (ret == 0) {
		printf("This is no Dual Shock controller??\n");
		return 1;
	}

	waitPadReady(port, slot);
	padSetMainMode(port, slot, PAD_MMODE_DUALSHOCK, PAD_MMODE_UNLOCK);

	return 1;
}

void setupPS2Pad() {
	int port;
	int slot = 0;
	int ret;

	for (port=0; port<2; port++) {
		if ((ret = padPortOpen(port, slot, padBuf[port])) == 0) {
			printf("padOpenPort failed: %d\n", ret);
		} else
			printf("padOpenPort success: %d\n", port);
		if (!initializePad(port, slot)) {
			printf("pad initalization failed!\n");
		} else
			printf("pad initialization success!\n");
	}
}
